<?php

	require_once realpath('.')."/DAVServer.php";
	define('DIRR','httpd/unix-directory');
    
	/**
	 * Perforce depot access using local p4 client
	 *
	 * @access public
	 */
class Depot_Filesystem extends HTTP_WebDAV_Server
{
	/**
	 * Root directory for WebDAV access
	 *
	 * @access private
	 * @var    string
	 */
	var $base = "";

	/** 
	 * Perforce Host connection parameters
	 *
	 * @access private
	 * @var    string
	 */
	var $depot_host = 'public.perforce.com';
	var $depot_port = '1666';
	var $depot_agent = '/usr/local/bin/p4';

	/**
	 * Perforce user
	 *
	 * @access private
	 * @var    string
	 */
	var $depot_user;

	/**
	 * Perforce password
	 *
	 * @access private
	 * @var    string
	 */
	var $depot_passwd;

	/**
	 * Serve a webdav request
	 *
	 * @access public
	 * @param  string  
	 */
	function ServeRequest($method) 
	{                           
	// let the base class do all the work
    parent::ServeRequest($method);		
 	}

	/**
	 *
	 * @access private
	 * @return bool true on successful authentication
	 */
	function _check_auth() 
	{
		if (empty($_SERVER['PHP_AUTH_USER'])) return false;
		// Authenticate with the Perforce server
		$this->depot_user = $_SERVER['PHP_AUTH_USER'];
		$this->depot_passwd = $_SERVER['PHP_AUTH_PW'];
// ************************
		$this->depot_user = 'daniel_sabsay';
		$this->depot_passwd = 'p4webdav';
		return true;
// ************************
		$r = $this->sp4cmd('user','-o','');
		$p = strpos($r,'Password:'); // bypass the commented password
		if (strpos(substr($r,$p+9),'Password:')) return true;
		$p4 = strpos($this->sp4cmd('-V','',''),'Perforce Software');
		if ($p4===false) echo 'Error: P4 client module not found at '.$this->depot_agent;
		return false;
	}
	
	/**
	 * PROPFIND method handler
	 *
	 * @param  array  general parameter passing array
	 * @param  array  return array for file properties
	 * @return bool   true on success
	 */
	function PROPFIND(&$options, &$files) 
	{
		$dirZ = array('path'=>': dir /');
		$fileZ = array('path'=>': depotFile /','size'=>': fileSize ');
		$depotZ = array('raw'=>'Depot ');
		$path = $options['path'];
		$p4path = '/'.$path;
		$files['files'] = array();
		if ($p4path=='//') {
			$group['dirs'] = array(array('path'=>''));
			$group['depots'] = $this->p4list('depots','',$depotZ);
		} else {
			if (substr($p4path,-1)=='/') {	
				$bpath = substr($p4path,0,strlen($p4path)-1);	
				$is_dir = 0<count($this->p4list('dirs',$bpath,$dirZ));
			} else {
				$is_dir = 0<count($this->p4list('dirs',$p4path,$dirZ));
				if ($is_dir) { $path .= '/'; $p4path .= '/'; }
				else { $group['files']=$this->p4list('fstat -Ol',$p4path,$fileZ); }
			}
			if (strpos($options['depth'],'noroot')===false) { // Omit self ?
				if ($is_dir) { $group['self']=array(array('path'=>$path)); }
				else { $group['files']=array(array('path'=>$path)); }
			}
			if ($is_dir&&!empty($options['depth'])) { // depth 0 = self only
				$group['dirs']=$this->p4list('dirs',$p4path.'*',$dirZ);			
				$group['files']=$this->p4list('fstat -Ol',$p4path.'*',$fileZ);
				//$group['config']=array($path.'/**my_configuration**');
			}
		}
		foreach ($group as $group_type=>$children) {
			foreach ($children as $child) {
				$props = array();
				switch ($group_type) {
				case 'depots':
					$child['path'] = '/'.substr($child['raw'],0,strpos($child['raw'],' '));
				case 'dirs':
					$child['path'] .= '/'; // sub-directories get trailing separator
				case 'self':
					$child['name'] = basename($child['path']);
					$child['res'] = 'collection';
					$child['type'] = DIRR;
					break;
				case 'files':
					$child['name'] = basename($child['path']);
					$child['type'] = 'text/plain';
				}
				array_push($props,$this->mkprop('getcontentlength',$child['size']));
				array_push($props,$this->mkprop('resourcetype',$child['res']));
				array_push($props,$this->mkprop('getcontenttype',$child['type']));
				array_push($props,$this->mkprop('displayname',$child['name']));
				array_push($props,$this->mkprop('creationdate','fakecreate'));
				array_push($props,$this->mkprop('getlastmodified','fakemodify'));
				array_push($files['files'],array('path'=>$child['path'],'props'=>$props));
			}
		}
	return;
	} 


	/**
	 * detect if a given program is found in the search PATH
	 *
	 * helper function used by _mimetype() to detect if the 
	 * external 'file' utility is available
	 *
	 * @param  string  program name
	 * @param  string  optional search path, defaults to $PATH
	 * @return bool    true if executable program found in path
	 */
	function _can_execute($name, $path = false) 
	{
			// path defaults to PATH from environment if not set
			if ($path === false) {
// *******                $path = getenv("PATH");
			}
	
			// check method depends on operating system
			if (!strncmp(PHP_OS, "WIN", 3)) {
					// on Windows an appropriate COM or EXE file needs to exist
					$exts = array(".exe", ".com");
					$check_fn = "file_exists";
			} else { 
					// anywhere else we look for an executable file of that name
					$exts = array("");
					$check_fn = "is_executable";
			}
			
			// now check the directories in the path for the program
			foreach (explode(PATH_SEPARATOR, $path) as $dir) {
					// skip invalid path entries
					if (!file_exists($dir)) continue;
					if (!is_dir($dir)) continue;

					// and now look for the file
					foreach ($exts as $ext) {
							if ($check_fn("$dir/$name".$ext)) return true;
					}
			}

			return false;
	}
     
	/**
	 * try to detect the mime type of a file
	 *
	 * @param  string  file path
	 * @return string  guessed mime type
	 */
	function _mimetype($fspath) 
	{
		if (@is_dir($fspath)) {
			// directories are easy
			return "httpd/unix-directory"; 
// ********         } else if (@function_exists("mime_content_type")) {
					// use mime magic extension if available
// ********            $mime_type = mime_content_type($fspath);
			} else if ($this->_can_execute("file")) {
				// it looks like we have a 'file' command, 
				// lets see it it does have mime support
				$fp = popen("file -i '".$fspath."' 2>/dev/null", "r");
				$reply = fgets($fp);
				pclose($fp);
					
				// popen will not return an error if the binary was not found
				// and find may not have mime support using "-i"
				// so we test the format of the returned string 
				
				// the reply begins with the requested filename
				if (!strncmp($reply, "$fspath: ", strlen($fspath)+2)) {                     
					$reply = substr($reply, strlen($fspath)+2);
					// followed by the mime type (maybe including options)
					if (ereg("^[[:alnum:]_-]+/[[:alnum:]_-]+;?.*", $reply, $matches)) {
							$mime_type = $matches[0];
					}
				}
			} 
			
		if (empty($mime_type)) {
		// Fallback solution: try to guess the type by the file extension
		// TODO: add more ...
		// TODO: it has been suggested to delegate mimetype detection 
		//       to apache but this has at least three issues:
		//       - works only with apache
		//       - needs file to be within the document tree
		//       - requires apache mod_magic 
		// TODO: can we use the registry for this on Windows?
		//       OTOH if the server is Windos the clients are likely to 
		//       be Windows, too, and tend do ignore the Content-Type
		//       anyway (overriding it with information taken from
		//       the registry)
		// TODO: have a seperate PEAR class for mimetype detection?
			switch (strtolower(strrchr(basename($fspath), "."))) {
			case ".html":
					$mime_type = "text/html";
					break;
			case ".gif":
					$mime_type = "image/gif";
					break;
			case ".jpg":
					$mime_type = "image/jpeg";
					break;
			case ".png":
					$mime_type = "image/png";
					break;
			case ".txt":
					$mime_type = "text/plain";
					break;
			case ".xml":
					$mime_type = "text/xml";
					break;
			case ".php":
					$mime_type = "text/php";
					break;
			default: 
					$mime_type = "application/octet-stream";
					break;
			}
		}
		
		return $mime_type;
	}

	/**
	 * GET method handler
	 * 
	 * @param  array  parameter passing array
	 * @return bool   true on success
	 */
	function GET(&$options) 
	{
		$path = $options['path'];
		$p4path = '/'.$path;
		$exists = 0<$this->p4list('files',$p4path,array(': depotFile /'));
		if (!$exists) return false;
		$file = $this->base.uniqid('p4'); // Generate temporary filename
		$this->sp4cmd('print -q -o '.$file,$p4path,''); // read data from depot
		$options['mimetype'] = 'text/html';
		$options['size'] = filesize($file);
		$options['stream'] = fopen($file,'r'); // Open stream to temp file
		$options['delete_path'] = $file; // Mark temp for disposal
		header('Accept-Ranges: bytes');
		header('ETag: "7b995d-4e30-3cf7d0b4"');
		return true;
	}
	
	/**
	 * PUT method handler
	 * 
	 * @param  array  parameter passing array
	 * @return bool   true on success
	 */
	function PUT(&$options) 
	{
		$path = $options['path'];
		$p4path = '/'.$path;
		$p4dir = '/'.dirname($path).'/';
		// MUST guard against directory name instead of full path
		$node = basename($path);
		// Create a temporary private directory on the server
		$uid = uniqid('');
		$ws_id = 'W'.$uid;
		$dir = $this->base.'D'.$uid; // temp directory name
		$fail = !mkdir($dir);
		if (!$fail) {
			// Copy file contents from the user's request
			$fr = $options['stream'];
			$fail = !$fr;
		}
		if (!$fail) {
			$file = $dir.'/'.$node;
			$fw = fopen($file,'w');
			$fail = !$fw;
		}
		if (!$fail) {
			while (!feof($fr)) fwrite($fw,fread($fr,4096));
			fclose($fr); fclose($fw);
			// Tell the Perforce client to copy the file to the depot
			$view  = 'Client:  '.$ws_id."\n\n";
			$view .= 'Owner:  '.$this->depot_user."\n\n";
			$view .= 'Options:  allwrite noclobber nocompress unlocked nomodtime normdir'."\n\n";
			$view .= 'LineEnd:  local'."\n\n";
			$view .= 'Root:  '.$dir."\n\n";
			$view .= 'View:  '."\n";
			$view .= '  '.$p4dir.'...  //'.$ws_id.'/...';  // needs quotes??
			$x = shell_exec('echo \''.$view.'\' | '.$this->p4cmd('client -i ','',''));
			$fail = strpos($x,'saved')===false;
		}
		if (!$fail) {
			$x = $this->sp4cmd('flush','','-c '.$ws_id);
			$fail = strpos($x,'added as')===false;
		}
		if (!$fail) {
			$change  = 'Change: new'."\n";
			$change .= 'Description:'."\n";
			$change .= ' Auto-submit from p4DAV/'.$this->depot_user."\n";
			$change .= 'Files:'."\n";
			$change .= '  '.$p4path;
			$exists = 0<count($this->p4list('files','"'.$p4path.'"',array(': depotFile /')));	
			$update = ($exists)? 'edit':'add';
			$x = $this->sp4cmd($update,'"'.$file.'"','-c '.$ws_id);
			$fail = strpos($x,'opened for')===false;
		}	
		if (!$fail) {
			$CR=shell_exec('echo \''.$change.'\' | '.$this->p4cmd('change -i','','-c '.$ws_id));
			$fail = strpos($CR,'created with')===false;
		}
		if (!$fail) {	
			$change_id = substr(substr($CR,0,strpos($CR,' created')),7);
			$x = $this->sp4cmd('submit','-c '.$change_id,'-c '.$ws_id);
			$fail = strpos($x,'submitted')===false;
		}
		if (!$fail) {
			$x = $this->sp4cmd('client','-d -f '.$ws_id,'');
			$fail = strpos($x,'deleted')===false;
		}	
		unlink($file);
		rmdir($dir);
//		$fw=fopen($dir.'/x1','w');fwrite($fw,$x);fclose($fw);
		return !$fail;
	}

	/**
	 * MKCOL method handler
	 *
	 * @param  array  general parameter passing array
	 * @return bool   true on success
	 */
	function MKCOL($options) 
	{           
			$path = $this->base . $options["path"];
			$parent = dirname($path);
			$name = basename($path);

			if(!file_exists($parent)) {
					return "409 Conflict";
			}

			if(!is_dir($parent)) {
					return "403 Forbidden";
			}

			if( file_exists($parent."/".$name) ) {
					return "405 Method not allowed";
			}

			if(!empty($_SERVER["CONTENT_LENGTH"])) { // no body parsing yet
					return "415 Unsupported media type";
			}
			
			$stat = mkdir ($parent."/".$name,0777);
			if(!$stat) {
					return "403 Forbidden";                 
			}

			return ("201 Created");
	}

	/**
	 * DELETE method handler
	 *
	 * @param  array  general parameter passing array
	 * @return bool   true on success
	 */
		/**
		 * DELETE method handler
		 *
		 * @param  array  general parameter passing array
		 * @return bool   true on success
		 */
		function DELETE($options) 
		{
				$path = $this->base . $options["path"];

				if(!file_exists($path)) return "404 Not found";

				if (is_dir($path)) {
						system("rm -rf $path");
				} else {
						unlink ($path);
				}
				return "204 No Content";
		}

	/**
	 * MOVE method handler
	 *
	 * @param  array  general parameter passing array
	 * @return bool   true on success
	 */
	function MOVE($options) 
	{
			return $this->copy($options, true);
	}

	/**
	 * COPY method handler
	 *
	 * @param  array  general parameter passing array
	 * @return bool   true on success
	 */
	function COPY($options, $del=false) 
	{
			// TODO Property updates still broken (Litmus should detect this?)

			
			if(!empty($_SERVER["CONTENT_LENGTH"])) { // no body parsing yet
					return "415 Unsupported media type";
			}

			// no copying to different WebDAV Servers yet
			if(isset($options["dest_url"])) {
					return "502 bad gateway";
			}

			$source = $this->base . $options["path"];
			if(!file_exists($source)) return "404 Not found";

			$dest = $this->base . $options["dest"];

			$new = !file_exists($dest);
			$existing_col = false;

			if(!$new) {
					if($del && is_dir($dest)) {
							if(!$options["overwrite"]) {
									return "412 precondition failed";
							}
							$dest .= basename($source);
							if(file_exists($dest.basename($source))) {
									$options["dest"] .= basename($source);
							} else {
									$new = true;
									$existing_col = true;
							}
					}
			}

			if(!$new) {
					if($options["overwrite"]) {
							$stat = $this->delete(array("path" => $options["dest"]));
							if($stat{0} != "2") return $stat; 
					} else {                
							return "412 precondition failed";
					}
			}

			if (is_dir($source)) {
					// RFC 2518 Section 9.2, last paragraph
					if ($options["depth"] != "infinity") {
							error_log("---- ".$options["depth"]);
							return "400 Bad request";
					}
					system(escapeshellcmd("cp -R ".escapeshellarg($source) ." " .  escapeshellarg($dest)));

					if($del) {
							system(escapeshellcmd("rm -rf ".escapeshellarg($source)) );
					}
			} else {                
					if($del) {
							@unlink($dest);
							rename($source, $dest);
					} else {
							if(substr($dest,-1)=="/") $dest = substr($dest,0,-1);
							copy($source, $dest);
					}
			}

			return ($new && !$existing_col) ? "201 Created" : "204 No Content";         
	}

	/**
	 * PROPPATCH method handler
	 *
	 * @param  array  general parameter passing array
	 * @return bool   true on success
	 */
	function PROPPATCH(&$options) 
	{
			$msg = "";
			
			$path = $options["path"];
			
			$dir = dirname($path)."/";
			$base = basename($path);
			
			foreach($options["props"] as $key => $prop) {
				if($ns == "DAV:") {
					$options["props"][$key][$status] = "403 Forbidden";
				}
			}							
			return "";
	}

	/**
	 * LOCK method handler
	 *
	 * @param  array  general parameter passing array
	 * @return bool   true on success
	 */
	function LOCK(&$options) 
	{
			if(isset($options["update"])) { // Lock Update					
			}
			
			$options["timeout"] = time()+300; // 5min. hardcoded

			return "200 OK";
	}

	/**
	 * UNLOCK method handler
	 *
	 * @param  array  general parameter passing array
	 * @return bool   true on success
	 */
	function UNLOCK(&$options) 
	{

			return "200 OK";
	}

	/**
	 * checkLock() helper
	 *
	 * @param  string resource path to check for locks
	 * @return bool   true on success
	 */
	function checkLock($path) 
	{
			// TODO
			return false;
	}


	/**
	 * create database tables for property and lock storage
	 *
	 * @param  void
	 * @return bool   true on success
	 */
	function create_database() 
	{
			// TODO
			return false;
	}

	function p4list($type,$path,$filters) {
		$data_lines = split("\n",$this->sp4cmd($type,$path,'-s -ztag'));
		$responses = array();
		$filter_count = count($filters);
		$hits = 1;
		$matches = array();
		foreach ($data_lines as $item) {
			foreach ($filters as $key=>$pattern) {
				$skip = strlen($pattern);
				$pos = strpos($item,$pattern);
				if ($pos===false) continue;
				$hits++;
				$matches[$key]=rtrim(substr($item,$pos+$skip));
			}
			if ($hits > $filter_count) {$hits = 1; array_push($responses,$matches);}
		}
		return $responses;
	}
	function p4cmd($cmd,$path,$gopts) {
		$p4cmd  = $this->depot_agent.' '.$gopts;
		$p4cmd .=' -u '.$this->depot_user;
		$p4cmd .=' -P '.$this->depot_passwd;
		$p4cmd .=' -p '.$this->depot_host.':'.$this->depot_port;
		$p4cmd .=' '.$cmd.' '.$path.' 2>&1';
		return $p4cmd;
	}
	function sp4cmd($a,$b,$c) { return shell_exec($this->p4cmd($a,$b,$c)); }
}
?>